/* -------------------------------------------------------------------------- *
 *                                   OpenMM                                   *
 * -------------------------------------------------------------------------- *
 * This is part of the OpenMM molecular simulation toolkit originating from   *
 * Simbios, the NIH National Center for Physics-Based Simulation of           *
 * Biological Structures at Stanford, funded under the NIH Roadmap for        *
 * Medical Research, grant U54 GM072970. See https://simtk.org.               *
 *                                                                            *
 * Portions copyright (c) 2008-2013 Stanford University and the Authors.      *
 * Authors: Peter Eastman                                                     *
 * Contributors:                                                              *
 *                                                                            *
 * Permission is hereby granted, free of charge, to any person obtaining a    *
 * copy of this software and associated documentation files (the "Software"), *
 * to deal in the Software without restriction, including without limitation  *
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,   *
 * and/or sell copies of the Software, and to permit persons to whom the      *
 * Software is furnished to do so, subject to the following conditions:       *
 *                                                                            *
 * The above copyright notice and this permission notice shall be included in *
 * all copies or substantial portions of the Software.                        *
 *                                                                            *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR *
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,   *
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL    *
 * THE AUTHORS, CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,    *
 * DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR      *
 * OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE  *
 * USE OR OTHER DEALINGS IN THE SOFTWARE.                                     *
 * -------------------------------------------------------------------------- */

#include "openmm/RPMDIntegrator.h"
#include "openmm/Context.h"
#include "openmm/OpenMMException.h"
#include "openmm/internal/ContextImpl.h"
#include "openmm/RpmdKernels.h"
#include <cmath>
#include <ctime>
#include <string>

using namespace OpenMM;
using std::string;
using std::vector;

RPMDIntegrator::RPMDIntegrator(int numCopies, double temperature, double frictionCoeff, double stepSize) :
        numCopies(numCopies), forcesAreValid(false), hasSetPosition(false), hasSetVelocity(false), isFirstStep(true) {
    setTemperature(temperature);
    setFriction(frictionCoeff);
    setStepSize(stepSize);
    setConstraintTolerance(1e-4);
    setRandomNumberSeed((int) time(NULL));
}

void RPMDIntegrator::initialize(ContextImpl& contextRef) {
    if (owner != NULL && &contextRef.getOwner() != owner)
        throw OpenMMException("This Integrator is already bound to a context");
    if (contextRef.getSystem().getNumConstraints() > 0)
        throw OpenMMException("RPMDIntegrator cannot be used with Systems that include constraints");
    context = &contextRef;
    owner = &contextRef.getOwner();
    kernel = context->getPlatform().createKernel(IntegrateRPMDStepKernel::Name(), contextRef);
    kernel.getAs<IntegrateRPMDStepKernel>().initialize(contextRef.getSystem(), *this);
}

void RPMDIntegrator::cleanup() {
    kernel = Kernel();
}

void RPMDIntegrator::stateChanged(State::DataType changed) {
    forcesAreValid = false;
}

vector<string> RPMDIntegrator::getKernelNames() {
    std::vector<std::string> names;
    names.push_back(IntegrateRPMDStepKernel::Name());
    return names;
}

void RPMDIntegrator::setPositions(int copy, const vector<Vec3>& positions) {
    kernel.getAs<IntegrateRPMDStepKernel>().setPositions(copy, positions);
    hasSetPosition = true;
}

void RPMDIntegrator::setVelocities(int copy, const vector<Vec3>& velocities) {
    kernel.getAs<IntegrateRPMDStepKernel>().setVelocities(copy, velocities);
    hasSetVelocity = true;
}

State RPMDIntegrator::getState(int copy, int types, bool enforcePeriodicBox, int groups) {
    if (isFirstStep) {
        // Call setPositions() on the Context so it doesn't think the user is trying to
        // run a simulation without setting positions first.  These positions will
        // immediately get overwritten by the ones stored in this integrator.
        
        vector<Vec3> p(context->getSystem().getNumParticles(), Vec3());
        context->getOwner().setPositions(p);
        isFirstStep = false;
    }
    kernel.getAs<IntegrateRPMDStepKernel>().copyToContext(copy, *context);
    State state = context->getOwner().getState(types, enforcePeriodicBox && copy == 0, groups);
    if (enforcePeriodicBox && copy > 0 && (types&State::Positions) != 0) {
        // Apply periodic boundary conditions based on copy 0.  Otherwise, molecules might end
        // up in different places for different copies.
        
        kernel.getAs<IntegrateRPMDStepKernel>().copyToContext(0, *context);
        State state2 = context->getOwner().getState(State::Positions, false, groups);
        vector<Vec3> positions = state.getPositions();
        const vector<Vec3>& refPos = state2.getPositions();
        const vector<vector<int> >& molecules = context->getMolecules();
        Vec3 periodicBoxSize[3];
        state2.getPeriodicBoxVectors(periodicBoxSize[0], periodicBoxSize[1], periodicBoxSize[2]);
        for (int i = 0; i < (int) molecules.size(); i++) {
            // Find the molecule center.

            Vec3 center;
            for (int j = 0; j < (int) molecules[i].size(); j++)
                center += refPos[molecules[i][j]];
            center *= 1.0/molecules[i].size();

            // Find the displacement to move it into the first periodic box.

            int xcell = (int) floor(center[0]/periodicBoxSize[0][0]);
            int ycell = (int) floor(center[1]/periodicBoxSize[1][1]);
            int zcell = (int) floor(center[2]/periodicBoxSize[2][2]);
            double dx = xcell*periodicBoxSize[0][0];
            double dy = ycell*periodicBoxSize[1][1];
            double dz = zcell*periodicBoxSize[2][2];

            // Translate all the particles in the molecule.

            for (int j = 0; j < (int) molecules[i].size(); j++) {
                Vec3& pos = positions[molecules[i][j]];
                pos[0] -= dx;
                pos[1] -= dy;
                pos[2] -= dz;
            }
        }
        
        // Construct the new State.
        
        State::StateBuilder builder(state.getTime());
        builder.setPositions(positions);
        builder.setPeriodicBoxVectors(periodicBoxSize[0], periodicBoxSize[1], periodicBoxSize[2]);
        if (types&State::Velocities)
            builder.setVelocities(state.getVelocities());
        if (types&State::Forces)
            builder.setForces(state.getForces());
        if (types&State::Parameters)
            builder.setParameters(state.getParameters());
        if (types&State::Energy)
            builder.setEnergy(state.getKineticEnergy(), state.getPotentialEnergy());
        state = builder.getState();
    }
    return state;
}

double RPMDIntegrator::computeKineticEnergy() {
    return kernel.getAs<IntegrateRPMDStepKernel>().computeKineticEnergy(*context, *this);
}

void RPMDIntegrator::step(int steps) {
    if (!hasSetPosition) {
        // Initialize the positions from the context.
        
        State s = context->getOwner().getState(State::Positions);
        for (int i = 0; i < numCopies; i++)
            setPositions(i, s.getPositions());
    }
    if (!hasSetVelocity) {
        // Initialize the velocities from the context.
        
        State s = context->getOwner().getState(State::Velocities);
        for (int i = 0; i < numCopies; i++)
            setVelocities(i, s.getVelocities());
    }
    if (isFirstStep) {
        // Call setPositions() on the Context so it doesn't think the user is trying to
        // run a simulation without setting positions first.  These positions will
        // immediately get overwritten by the ones stored in this integrator.
        
        vector<Vec3> p(context->getSystem().getNumParticles(), Vec3());
        context->getOwner().setPositions(p);
        isFirstStep = false;
    }
    for (int i = 0; i < steps; ++i) {
        kernel.getAs<IntegrateRPMDStepKernel>().execute(*context, *this, forcesAreValid);
        forcesAreValid = true;
    }
}
